import java.util.Random;
/**
 * Simple AI to determine the best course of attack/Item use/Monster switch
 * based on the two battling Monsters.
 * 
 * @see Monster
 * @see Battle
 * @author Samuel Baumgardner
 */
public class BattleAI
{
	/**
	 * Recommends the best Attack corresponding for the Monster
	 * 
	 * @param attacker the Monster attacking
	 * @param opponent the Monster defending
	 * @return the recommended Attack
	 */
	public static Attack recommendMonsterAttack(Monster attacker, Monster opponent)
	{
		//requires: attacker and opponent be non-null and initialized
		//that myMonster's attacks are sorted from lowest to highest
		//effects: returns the best attack for attacker.

		Monster myMonster = attacker;
		Monster mangyBastard = opponent;

		//Strategy one: If my opponent is weak against me then sock it to him.
		if(isWeakAgainst(myMonster.getType(),mangyBastard.getType()))
			return myMonster.getAttack(myMonster.getAttacks().length-1);

		//Strategy two: If I am weaker than my opponent tip toe thru the house and try not to disturb.
		if(isWeakAgainst(mangyBastard.getType(),myMonster.getType()))
			return myMonster.getAttack(0);

		//Strategy three: return a reasonable attack
		//if the random number is greater than the current level then choose the highest
		//attack possible
		//if the random number is less than the current level then choose a random attack that is weaker than the highest
		//attack possible


		int attack;
		Random random = new Random();
		random.setSeed(System.currentTimeMillis());
		attack = random.nextInt(myMonster.getMaxLevel());
		if(attack <= (myMonster.getLevel()-1))
			return myMonster.getAttack(myMonster.getNumberOfAttacks()-1);
		attack = random.nextInt(myMonster.getNumberOfAttacks()-1);
		return myMonster.getAttack(attack);
	}

	/**
	 * Determines if the itemType is available, ie is in the People's
	 * Item collection.
	 * 
	 * @param itemType the brand of the Item to search for
	 * @param itemSubType the subBrand of the Item to search for
	 * @param items the String array of items the People is carrying
	 * @param itemContainer the People to search in.
	 * @return the Item if it is owned, else null.
	 * @see Item
	 * @see People
	 */
	private static Item itemLookUp(int itemType, int itemSubType, String[] items, People itemContainer)
	{
		for(int i = 0;i < items.length;i++)
		{
			if(items[i] != null)
			{
				Item item = itemContainer.getItem(items[i]);
				if(item.getBrand() == itemType)
				{
					if(item.getSubBrand() == itemSubType)
						return item;
				}
			}
		}
		return null;
	}
	/**
	 * Recommends the best Attack corresponding for the People against the Player
	 * 
	 * @param attacker the People attacking
	 * @param opponent the Player defending
	 * @return the recommended Attack
	 * @exception ItemChangeException if the best course of action is to use an Item
	 * @exception MonsterChangeException if the best course of action is to switch Monsters
	 * @see Monster
	 * @see People
	 * @see Player
	 * @see Item
	 */
	public static Attack recommendPlayerAttack(People attacker, Player opponent) throws ItemChangeException, MonsterChangeException
	{
		Monster opp = opponent.getMonster(opponent.getActiveMonster());
		Monster att = attacker.getMonster(attacker.getActiveMonster());


		//Strategy 1: If health points are low < 20% use max potion to increase health
		//Strategy 2: If defense is low < 20% use potion to increase defense
		//Strategy 3: If attack is low < 20% use potion to increase attack

			String[] items = attacker.getItems();

			PotionItem potion = null;

			if((100 * (att.getDefense()/att.getMaxDefense())) < 20)
			{
				potion = (PotionItem)itemLookUp(Item.POTION, PotionItem.DEFUP, attacker.getItems(), attacker);
				if(potion != null)
					throw new ItemChangeException(potion);
			}
			
			if((100 * (att.getAttack()/att.getMaxAttack())) < 20)
			{
				potion = (PotionItem)itemLookUp(Item.POTION, PotionItem.ATTUP, attacker.getItems(), attacker);
				if(potion != null)
					throw new ItemChangeException(potion);
			}

			if((100 * (att.getHP()/att.getMaxHP())) < 20)
			{
				potion = (PotionItem)itemLookUp(Item.POTION, PotionItem.HPUPFULL, attacker.getItems(), attacker);
				if(potion != null)
					throw new ItemChangeException(potion);

				potion = (PotionItem)itemLookUp(Item.POTION, PotionItem.HPUPHALF, attacker.getItems(), attacker);
				if(potion != null)
					throw new ItemChangeException(potion);

				potion = (PotionItem)itemLookUp(Item.POTION, PotionItem.HPUPTHIRD, attacker.getItems(), attacker);
				if(potion != null)
					throw new ItemChangeException(potion);
			}

		if( isWeakAgainst(opp.getKind(), att.getKind()) )
		{
			for(int i = 0; i < attacker.getNumberOfMonsters(); i++)
			{
				if( !isWeakAgainst(opp.getKind(), attacker.getMonster(i).getKind()) )
					throw new MonsterChangeException(attacker.getActiveMonster(), i);
			}
		}
		return recommendMonsterAttack(att, opp);
	}

	/**
	 * Determine if a Monster type is weak against another.
	 * 
	 * @param plt the one to determine based on, ie the attacker
	 * @param opt the one to determine against, ie the defender
	 * @return true if opt is weak against plt, false othrwise.
	 */
	private static boolean isWeakAgainst(int plt, int opt) // If opt < plt
	{
		switch(plt)
		{
			case Monster.GAS:
			{
				if(opt == Monster.ELECTRIC || opt == Monster.EARTH)
					return true;
				return false;
			}
			case Monster.FIRE:
			{
				if(opt == Monster.GAS || opt == Monster.ELECTRIC)
					return true;
				return false;
			}
			case Monster.WATER:
			{
				if(opt == Monster.GAS || opt == Monster.FIRE)
					return true;
				return false;
			}
			case Monster.EARTH:
			{
				if(opt == Monster.WATER || opt == Monster.FIRE)
					return true;
				return false;
			}
			case Monster.ELECTRIC:
			{
				if(opt == Monster.EARTH || opt == Monster.WATER)
					return true;
				return false;
			}
		}
		return false;
	}

}