import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.awt.geom.*;
import java.awt.image.*;

/**
 * Class for displaying the map and handling the Player Movements.  Fires Events for
 * battles and redirects
 * 
 * @see Player
 * @see People
 * @see MapObject
 * @see Redirect
 * @see RedirectEvent
 * @see Game
 * @see BattleEvent
 */
public class Grid extends Panel 
{
	/**
	 * The default number of Rows/COlumns
	 */
	public static final int DEFAULTNUM = 16;

	/**
	 * Left direction
	 */
	public static final int LEFT = 0;
	/**
	 * Right direction
	 */
	public static final int RIGHT = 1;
	/**
	 * Up direction
	 */
	public static final int UP = 2;
	/**
	 * Down direction
	 */
	public static final int DOWN = 3;

	/**
	 * The number of rows
	 */
	protected int numi; // Num Rows
	/**
	 * The number of columns
	 */
	protected int numj; // Num Columns

	/**
	 * The position of the player, the row/i
	 */
	protected int pi; // Player position i
	/**
	 * The position of the player, the column/j
	 */
	protected int pj; // Player position j

	/**
	 * The BattleListeners
	 */
	private Vector battleListeners;
	/**
	 * The RedirectEventListeners
	 */
	private Vector redirectListeners;

	/**
	 * This
	 */
	MapObject mp[][];
	/**
	 * Default backdrop
	 */
	MapObject def;
	/**
	 * The Monsters allowed on this
	 */
	Vector monAllow;
	/**
	 * The Monster level for wild Monsters
	 */
	int monsterLevel;
	/**
	 * is there a battle
	 */
	boolean battle;

	/**
	 * Needed for calls to paintAll
	 */
	Graphics g; // Needed for call to repaint

	/**
	 * Construct a new Grid of size Grid.DEFAULTNUM
	 */
	public Grid()
	{
		mp = new MapObject[DEFAULTNUM][DEFAULTNUM];
		numi = DEFAULTNUM;
		numj = DEFAULTNUM;
		monAllow = new Vector();
		monsterLevel = 10;
		setLayout(new GridLayout(numi, numj, 0, 0));
		battle = false;
		enableEvents( AWTEvent.KEY_EVENT_MASK);
	}

	/**
	 * Gets the number of rows.
	 * 
	 * @return the Number of rows
	 */
	public int getNumRows() { return numi; }
	/**
	 * Gets the number of columns
	 * 
	 * @return the number of columns
	 */
	public int getNumColumns() { return numj; }

	/**
	 * Determines if the Location specified is a valid on this.
	 * 
	 * @param i the row
	 * @param j the column
	 * @return true iff i & j are in this ie less than Grid.DEFAULTNUM, and greater than 0
	 */
	public synchronized boolean isValidLocation(int i, int j)
	{
		if(i >= 0 && i < numi && j >= 0 && j < numj)
			return true;
		return false;
	}

	/**
	 * Determines if the Location specified can be moved into
	 * 
	 * @param i the row
	 * @param j the column
	 * @return truee iff can be moved into, ie if isValidLocation, and canBeEntered
	 * @see MapObject
	 * @see Grass
	 * @see Floor
	 */
	public synchronized boolean canMove(int i, int j)
	{
		if(!isValidLocation(i, j))
			return false;
		if(mp[i][j] == null)
			return true;
		return mp[i][j].canBeEntered();
	}
	/**
	 * Sets the default backdrop to display.
	 * 
	 * @param m the MapObject to have as the backdrop
	 */
	public synchronized void setDefault(MapObject m)
	{
		if(m == null)
			return;
		if(m.canBeEntered())
		{
			this.def = m;
		}
	}
	/**
	 * Returns the preferred size of this.
	 * Needs be larger or equal to the size of the Applet
	 * 
	 * @return the Dimension of this
	 * @see Game
	 */
	public Dimension getPerferredSize()
	{
		return new Dimension(640, 480);
	}
	/**
	 * Sets the Player of this.
	 * 
	 * @param p the new Player of this
	 */
	public synchronized void setPlayer(Player p)
	{
		if(p != null)
		{
			try 
			{
				removeObject(pi, pj);
				setObjectAt(p, pi, pj);
			} 
			catch(Exception e) { }
		}
	}
	/**
	 * Sets the object at the Location specified
	 * 
	 * @param mp the MapObject to set at i, j
	 * @param i the row
	 * @param j the column
	 * @exception InvalidLocationException iff !isValidException(i, j)
	 * @exception NonEmptyException iff !canMove(i, j)
	 */
	public synchronized void setObjectAt(MapObject mp, int i, int j) throws InvalidLocationException, NonEmptyException
	{
		if(isValidLocation(i, j))
		{
			if(canMove(i, j))
			{
				try
				{
					remove(this.mp[i][j]);
				}
				catch(Exception e) { }
				this.mp[i][j] = mp;
				add(this.mp[i][j], numi*i+j);
				if(mp.getType() == MapObject.PLAYER)
				{
					pi = i; pj = j;
				}
			}
			else
				throw new NonEmptyException("(" + i + "," + j + ") cannot be moved into");
		}
		else
			throw new InvalidLocationException("Invalid Location (" + i + "," + j + ")");
	}
	/**
	 * Removes and returns the MapObject at the specified Location
	 * 
	 * @param i the row
	 * @param j the column
	 * @exception InvalidLocationException iff !isValidException(i, j)
	 */
	public synchronized MapObject removeObject(int i, int j) throws InvalidLocationException
	{
		if(isValidLocation(i, j))
		{
			remove(mp[i][j]);
			MapObject mo = mp[i][j];
			mp[i][j] = def.getClone();
			add(mp[i][j], numi*i+j);
			return mo;
		}
		else
			throw new InvalidLocationException("Invalid Location (" + i + "," + j + ")");
	}
	/**
	 * Remove All Objects from the Grid.
	 */
	public synchronized void removeAllObjects()
	{
		Player player = (Player)mp[pi][pj];
		for(int i = 0; i < numi; i++)
		{
			for(int j = 0; j < numj; j++)
			{
				mp[i][j] = def.getClone();
			}
		}
		mp[0][0] = player;
	}

	/**
	 * Gets the MapObject at i, j.
	 * 
	 * @param i the row
	 * @param j the columne
	 * @exception InvalidLocationException iff !isValidException(i, j)
	 * @exception EmptyLocationException if the Location only contains the default MapObject
	 */
	public synchronized MapObject mapObjectAt(int i, int j) throws EmptyLocationException, InvalidLocationException
	{
		if(battle)
			return null;
		if(isValidLocation(i, j))
		{
			if( !(mp[i][j].canBeEntered()) )
			{
				return mp[i][j];
			}
			else
			{
				throw new EmptyLocationException("("+ i + "," + j + " contains no usable Object");
			}
		}
		else
			throw new InvalidLocationException("("+ i + "," + j + " is not a valid location");
	}

	/**
	 * Gets the Location of the specified MapObject, uses equals(Object o).
	 * Has a O(n*n) run time.
	 * 
	 * @param mp the MapObject to search for.
	 */
	public Location locationOf(MapObject mp) throws InvalidLocationException
	{
		for(int i = 0; i < numi; i++)
		{
			for(int j = 0; j < numj; j++)
			{
				if(this.mp[i][j] == mp)
					return new Location(i, j);
			}
		}
		throw new InvalidLocationException("MapObject not found on Grid");
	}
	/**
	 * Returns the MapObject in the direction specified within the distance specified from the i, j specified.
	 * 
	 * @param i the row
	 * @param j the column
	 * @param dir the direction as specified above
	 * @param distance the distance to look
	 * @exception InvalidLocationException iff !isValidException(i, j)
	 */
	public synchronized MapObject isPeopleInDirection(int i, int j, int dir, int distance) throws InvalidLocationException
	{
		if(battle)
			return null;
		if(isValidLocation(i, j))
		{
			int d = 1;
			if(dir == LEFT)
			{
				for(--j; j < numj && d <= distance && j >= 0; j--, d++)
				{
					if(mp[i][j].getType() == MapObject.PLAYER || mp[i][j].getType() == MapObject.PERSON)
					{
						return mp[i][j];
					}
					else if(!(mp[i][j].canBeEntered())) // can't see over rocks, etc
					{
						return null;
					}
				}
			}
			else if(dir == RIGHT)
			{
				for(++j; j < numj && d <= distance && j >=0; j++, d++)
				{
					if(mp[i][j].getType() == MapObject.PLAYER || mp[i][j].getType() == MapObject.PERSON)
					{
						return mp[i][j];
					}
					else if(!(mp[i][j].canBeEntered())) // can't see over rocks, etc
					{
						return null;
					}
				}
			}
			else if(dir == UP)
			{
				for(--i; i < numi && d <= distance && i >= 0; i--, d++)
				{
					if(mp[i][j].getType() == MapObject.PLAYER || mp[i][j].getType() == MapObject.PERSON)
					{
						return mp[i][j];
					}
					else if(!(mp[i][j].canBeEntered())) // can't see over rocks, etc
					{
						return null;
					}
				}
			}
			else if(dir == DOWN)
			{
				for(++i; i < numi && d <= distance && i >= 0; i++, d++)
				{
					if(mp[i][j].getType() == MapObject.PLAYER || mp[i][j].getType() == MapObject.PERSON)
					{
						return mp[i][j];
					}
					else if(!(mp[i][j].canBeEntered())) // can't see over rocks, etc
					{
						return null;
					}
				}
			}
			return null;
		}
		else
			throw new InvalidLocationException("("+ i + "," + j + " is not a valid location");
	}
	/**
	 * Gets the Player of this.
	 * 
	 * @return the Player of this
	 */
	public synchronized Player getPlayer() { return (Player)mp[pi][pj]; }
	/**
	 * Moves the Player in the direction specified
	 * 
	 * @param direction the direction to move the player, as specified above
	 */
	public synchronized void movePlayerTo(int direction) throws NonEmptyException, InvalidLocationException
	{
		battle = false;
		switch(direction)
		{
			case LEFT:
			{
				if(pj-1 >= 0)
				{
					if(mp[pi][pj-1].getType() == MapObject.REDIRECT)
					{
						if(((Player)mp[pi][pj]).getBadgeCount() >= ((Redirect)mp[pi][pj-1]).getNumberBadgesRequired())
							fireRedirectEvent((Redirect)mp[pi][pj-1]);
						return;
					}
					else if(!canMove(pi, pj-1))
						return;
					Player player = (Player)removeObject(pi, pj);
					pj--;
					if(mp[pi][pj].getType() == MapObject.MONSTER)
					{
						Monster m = (Monster)removeObject(pi, pj);
						setObjectAt(player, pi, pj);
						beginBattleWithPlayer(m);
					}
					else
					{
						setObjectAt(player, pi, pj);
					}
				}
				else
					throw new NonEmptyException("(" + (pi-1) + "," + pj + ") is cannot be moved into");
				break;
			}
			case RIGHT:
			{
				if(pj+1 < numj)
				{
					if(mp[pi][pj+1].getType() == MapObject.REDIRECT)
					{
						if(((Player)mp[pi][pj]).getBadgeCount() >= ((Redirect)mp[pi][pj+1]).getNumberBadgesRequired())
							fireRedirectEvent((Redirect)mp[pi][pj+1]);
						return;
					}
					else if(!canMove(pi, pj+1))
						return;
					Player player = (Player)removeObject(pi, pj);
					pj++;
					if(mp[pi][pj].getType() == MapObject.MONSTER)
					{
						Monster m = (Monster)removeObject(pi, pj);
						setObjectAt(player, pi, pj);
						beginBattleWithPlayer(m);
					}
					else
					{
						setObjectAt(player, pi, pj);
					}
				}
				else
					throw new NonEmptyException("(" + (pi+1) + "," + pj + ") is cannot be moved into");
				break;
			}
			case UP:
			{
				if(pi-1 >= 0)
				{
					if(mp[pi-1][pj].getType() == MapObject.REDIRECT)
					{
						if(((Player)mp[pi][pj]).getBadgeCount() >= ((Redirect)mp[pi-1][pj]).getNumberBadgesRequired())
							fireRedirectEvent((Redirect)mp[pi-1][pj]);
						return;
					}
					else if(!canMove(pi-1, pj))
						return;
					Player player = (Player)removeObject(pi, pj);
					pi--;
					if(mp[pi][pj].getType() == MapObject.MONSTER)
					{
						Monster m = (Monster)removeObject(pi, pj);
						setObjectAt(player, pi, pj);
						beginBattleWithPlayer(m);
					}
					else
					{
						setObjectAt(player, pi, pj);
					}
				}
				else
					throw new NonEmptyException("(" + pi + "," + (pj-1) + ") is cannot be moved into");
				break;
			}
			case DOWN:
			{
				if(pi+1 < numj)
				{
					if(mp[pi+1][pj].getType() == MapObject.REDIRECT)
					{
						if(((Player)mp[pi][pj]).getBadgeCount() >= ((Redirect)mp[pi+1][pj]).getNumberBadgesRequired())
							fireRedirectEvent((Redirect)mp[pi+1][pj]);
						return;
					}
					else if(!canMove(pi+1, pj))
						return;
					Player player = (Player)removeObject(pi, pj);
					pi++;
					if(mp[pi][pj].getType() == MapObject.MONSTER)
					{
						Monster m = (Monster)removeObject(pi, pj);
						setObjectAt(player, pi, pj);
						beginBattleWithPlayer(m);
					}
					else
					{
						setObjectAt(player, pi, pj);
					}
				}
				else
					throw new NonEmptyException("(" + pi + "," + (pj+1) + ") is cannot be moved into");
				break;
			}
		}
	}

	/**
	 * Moves the MapObject at i, j to ni, nj.
	 * 
	 * @param i the row originally at
	 * @param j the column origanally at
	 * @param ni the new row
	 * @param nj the new column
	 * @exception InvalidLocationException iff !isValidException(i, j)
	 * @exception NonEmptyException iff !canMove(i, j)
	 * @exception EmptyLocationException Will Not Be Thrown.
	 */
	public synchronized void moveMapObjectTo(int i, int j, int ni, int nj) throws NonEmptyException, InvalidLocationException, EmptyLocationException
	{
		if(battle)
			return;
		if(isValidLocation(i, j))
		{
			if(isValidLocation(ni, nj))
			{
				if(canMove(ni, nj))
				{
					MapObject mp = removeObject(i, j);
					if(mp.getType() == MapObject.PERSON)
						setObjectAt(mp, ni, nj);
				}
				else
					throw new NonEmptyException("(" + (i) + "," + j + ") is cannot be moved into");
			}
		}
		else
			throw new InvalidLocationException("(" + i + "," + j + ") is not a valid location");
	}
	/**
	 * Handle the Events.  Overriden to handle the KeyEvents.
	 * 
	 * @param e the AWTEvent to process
	 */
	public synchronized void processEvent(AWTEvent e) 
	{
		if(e.getID() == KeyEvent.KEY_PRESSED)
		{
			int ti = pi, tj = pj;
			try
			{
				switch(((KeyEvent)e).getKeyCode())
				{
					case KeyEvent.VK_LEFT:
						{
						movePlayerTo(LEFT);
						((Player)mp[pi][pj]).setDirection(LEFT);
						//((Player)mp[pi][pj]).stopBattle();
						break;
					}
					case KeyEvent.VK_RIGHT:
						{
						movePlayerTo(RIGHT);
						((Player)mp[pi][pj]).setDirection(RIGHT);
						//((Player)mp[pi][pj]).stopBattle();
						break;
					}
					case KeyEvent.VK_UP:
						{
						movePlayerTo(UP);
						((Player)mp[pi][pj]).setDirection(UP);
						//((Player)mp[pi][pj]).stopBattle();
						break;
					}
					case KeyEvent.VK_DOWN:
						{
						movePlayerTo(DOWN);
						((Player)mp[pi][pj]).setDirection(DOWN);
						//((Player)mp[pi][pj]).stopBattle();
						break;
					}
				}
				repaint(ti, tj);
				repaint(pi, pj);
			}
			catch (Exception ile) { }
		}
		super.processEvent(e);
	}
	/**
	 * Allow wild Monsters of type m on this.
	 * 
	 * @param m the Monster.<type> to allow
	 */
	public void allowMonsterOnMap(int m)
	{
		if(m == Monster.ELECTRIC || m == Monster.EARTH || m == Monster.FIRE || m == Monster.WATER || m == Monster.GAS)
			monAllow.add(new Integer(m));
	}
	/**
	 * Sets the wild Monsters level for this
	 * 
	 * @param l the level of the wild Monsters
	 */
	public void setMonsterLevel(int l)
	{
		if(l > 1 && l < Monster.MAXLEVEL)
			monsterLevel = l;
	}
	/**
	 * Fills the Grid in a standard manner, useful for debugging.
	 */
	public void fillGrid() // Temporary until MAP implemented
	{
		battle = false;
		int wi, hi;
		wi = getWidth()/numj;
		hi = getHeight()/numi;
		def = new Grass();

		for(int i = 0; i < numi; i++)
		{
			for(int j = 0; j < numj; j++)
			{
				if(mp[i][j] == null)
				{
					if(i%5 == j%10)
						mp[i][j] = new Rock();
					else
						mp[i][j] = def.getClone();
					add(mp[i][j]);
				}
			}
		}
		try
		{
			removeObject(0, 0);
			People pl = new Player();
			setObjectAt(pl, 0, 0);
			pi = 0; pj = 0;

			removeObject(15, 15);
			pl = new Person(DOWN);
			setObjectAt(pl, 15, 15);
			((Person)pl).setGrid(this);

			removeObject(numi/2,numj/2);
			pl = new Person(LEFT);
			setObjectAt(pl, numi/2, numj/2);
			((Person)pl).setGrid(this);

			removeObject(numi/4,numj/4);
			pl = new Person(LEFT);
			setObjectAt(pl, numi/4, numj/4);
			((Person)pl).setGrid(this);

			removeObject(numi/5,numj/5);
			pl = new Person(LEFT);
			setObjectAt(pl, numi/5, numj/5);
			((Person)pl).setGrid(this);

			removeObject(numi/6,numj/6);
			pl = new Person(LEFT);
			setObjectAt(pl, numi/5, numj/5);
			((Person)pl).setGrid(this);
		}
		catch(Exception e) { }
	}
	/**
	 * Initialize the Grid, ie start the Peopl Moving, and fill this
	 * with wild Monsters
	 */
	public void init()
	{
		battle = false;
		try
		{
			Random r = new Random();
			r.setSeed(System.currentTimeMillis());
			for(int i = 0; i < DEFAULTNUM; i++)
			{
				for(int j = 0; j < DEFAULTNUM; j++)
				{
					if(mp[i][j].getType() == MapObject.PERSON)
					{
						((Person)mp[i][j]).setGrid(this);
					}
					else if(mp[i][j].getType() == MapObject.TRAINER)
					{
						((Trainer)mp[i][j]).setGrid(this);
					}
					else if(mp[i][j].getType() == MapObject.MONSTER)
					{
						removeObject(i, j);
					}
				}
			}
			if(monAllow.size() <= 0)
				return;

			// Adding 10 Monsters Randomly distrubuted
			Monster m;
			MapObject ma;
			int lev10, lev7, lev6;
			lev10 = monsterLevel*10;
			lev7 = monsterLevel*7;
			lev6 = monsterLevel*6;
			for(int i = 0, j = 0, s, t, u; i < 10; i++, j=0)
			{
				do
				{
					j++;
					s = r.nextInt(DEFAULTNUM);
					t = r.nextInt(DEFAULTNUM);
				} while( !(mp[s][t].canBeEntered()) && j < 5 );
				removeObject(s, t);
				ma = def.getClone();
				u = r.nextInt(monAllow.size());
				switch(((Integer)monAllow.get(u)).intValue())
				{
					case Monster.FIRE:
					{
						m = new FireMon((s%2 == 0 ? "MoPyro" : "Flamesaur"));
						m.setBehind(ma);
						
						m.setLevel(monsterLevel);
						m.setMaxHP(lev10);
						m.setHP(lev10);
						m.setDefense(lev7);
						m.setMaxDefense(lev7);
						m.setAttack(lev6);
						m.setMaxAttack(lev6);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(FireMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(FireMon.A4);

						setObjectAt(m,s, t);
						break;
					}
					case Monster.WATER:
					{
						m = new WaterMon((s%2 == 0 ? "Alligatrix" : "Sharkasaur"));
						m.setBehind(ma);

						m.setLevel(monsterLevel);
						m.setMaxHP(lev6);
						m.setHP(lev6);
						m.setDefense(lev10);
						m.setMaxDefense(lev10);
						m.setAttack(lev7);
						m.setMaxAttack(lev7);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(WaterMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(WaterMon.A4);

						setObjectAt(m,s, t);
						break;
					}
					case Monster.EARTH:
					{
						m = new EarthMon((s%2 == 0 ? "Rhono" : "Groundyx"));
						m.setBehind(ma);
						
						m.setLevel(monsterLevel);
						m.setMaxHP(lev10);
						m.setHP(lev10);
						m.setDefense(lev6);
						m.setMaxDefense(lev6);
						m.setAttack(lev7);
						m.setMaxAttack(lev7);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(EarthMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(EarthMon.A4);

						setObjectAt(m,s, t);
						break;
					}
					case Monster.GAS:
					{
						m = new GasMon((s%2 == 0 ? "Smelltor" : "Eurekasaur"));
						m.setBehind(ma);
						
						m.setLevel(monsterLevel);
						m.setMaxHP(lev7);
						m.setHP(lev7);
						m.setDefense(lev10);
						m.setMaxDefense(lev10);
						m.setAttack(lev6);
						m.setMaxAttack(lev6);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(GasMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(GasMon.A4);

						setObjectAt(m,s, t);
						break;
					}
					case Monster.ELECTRIC:
					{
						m = new ElectricMon((s%2 == 0 ? "Shockachu" : "Voltasaur"));
						m.setBehind(ma);
						
						m.setLevel(monsterLevel);
						m.setMaxHP(lev6);
						m.setHP(lev6);
						m.setDefense(lev7);
						m.setMaxDefense(lev7);
						m.setAttack(lev10);
						m.setMaxAttack(lev10);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(ElectricMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(ElectricMon.A4);

						setObjectAt(m,s, t);
						break;
					}
				}
			}
		} 
		catch(Exception e) { }
	}
	/**
	 * Removes the older wild Monsters, and add new ones randomly
	 * from th selection allowed.
	 */
	public void fillMonsters()
	{
		try
		{
			for(int i = 0; i < DEFAULTNUM; i++)
			{
				for(int j = 0; j < DEFAULTNUM; j++)
				{
					if(mp[i][j].getType() == MapObject.MONSTER)
					{
						removeObject(i, j);
					}
				}
			}
			Random r = new Random();
			r.setSeed(System.currentTimeMillis());
			battle = false;
			if(monAllow.size() <= 0)
				return;
			// Adding 10 Monsters Randomly distrubuted
			Monster m;
			MapObject ma;
			int lev10, lev7, lev6;
			lev10 = monsterLevel*10;
			lev7 = monsterLevel*7;
			lev6 = monsterLevel*6;
			for(int i = 0, s, t, u, j = 0; i < 10; i++, j = 0)
			{
				do
				{
					j++;
					s = r.nextInt(DEFAULTNUM);
					t = r.nextInt(DEFAULTNUM);
				} while( !(mp[s][t].canBeEntered()) && j < 5);
				removeObject(s, t);
				ma = def.getClone();
				u = r.nextInt(monAllow.size());
				switch(((Integer)monAllow.get(u)).intValue())
				{
					case Monster.FIRE:
						{
						m = new FireMon((s%2 == 0 ? "MoPyro" : "Flamesaur"));
						m.setBehind(ma);
						
						m.setLevel(monsterLevel);
						m.setMaxHP(lev10);
						m.setHP(lev10);
						m.setDefense(lev7);
						m.setMaxDefense(lev7);
						m.setAttack(lev6);
						m.setMaxAttack(lev6);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(FireMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(FireMon.A4);

						setObjectAt(m,s, t);
						break;
					}
					case Monster.WATER:
						{
						m = new WaterMon((s%2 == 0 ? "Alligatrix" : "Sharkasaur"));
						m.setBehind(ma);

						m.setLevel(monsterLevel);
						m.setMaxHP(lev6);
						m.setHP(lev6);
						m.setDefense(lev10);
						m.setMaxDefense(lev10);
						m.setAttack(lev7);
						m.setMaxAttack(lev7);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(WaterMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(WaterMon.A4);

						setObjectAt(m,s, t);
						break;
					}
					case Monster.EARTH:
						{
						m = new EarthMon((s%2 == 0 ? "Rhono" : "Groundyx"));
						m.setBehind(ma);
						
						m.setLevel(monsterLevel);
						m.setMaxHP(lev10);
						m.setHP(lev10);
						m.setDefense(lev6);
						m.setMaxDefense(lev6);
						m.setAttack(lev7);
						m.setMaxAttack(lev7);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(EarthMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(EarthMon.A4);

						setObjectAt(m,s, t);
						break;
					}
					case Monster.GAS:
						{
						m = new GasMon((s%2 == 0 ? "Smelltor" : "Eurekasaur"));
						m.setBehind(ma);
						
						m.setLevel(monsterLevel);
						m.setMaxHP(lev7);
						m.setHP(lev7);
						m.setDefense(lev10);
						m.setMaxDefense(lev10);
						m.setAttack(lev6);
						m.setMaxAttack(lev6);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(GasMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(GasMon.A4);

						setObjectAt(m,s, t);
						break;
					}
					case Monster.ELECTRIC:
						{
						m = new ElectricMon((s%2 == 0 ? "Shockachu" : "Voltasaur"));
						m.setBehind(ma);
						
						m.setLevel(monsterLevel);
						m.setMaxHP(lev6);
						m.setHP(lev6);
						m.setDefense(lev7);
						m.setMaxDefense(lev7);
						m.setAttack(lev10);
						m.setMaxAttack(lev10);
						m.setMaxExperience(lev10);
						m.setExperience(Math.abs(lev10-20));

						if(monsterLevel > 30)
							m.addAttack(ElectricMon.A3);
						if(monsterLevel >= 40)
							m.addAttack(ElectricMon.A4);

						setObjectAt(m,s, t);
						break;
					}
				}
			}
		}
		catch(Exception e) { }
	}
	/**
	 * Destroy all threads tied to this by setting their
	 * Grid Objects to null.
	 * 
	 * @see PersonAI
	 * @see Trainer
	 */
	public void destroy()
	{
		for(int i = 0; i < DEFAULTNUM; i++)
		{
			for(int j = 0; j < DEFAULTNUM; j++)
			{
				if(mp[i][j].getType() == MapObject.PERSON)
				{
					((Person)mp[i][j]).setGrid(null); // Kill AI
				}
				else if(mp[i][j].getType() == MapObject.TRAINER)
				{
					((Trainer)mp[i][j]).setGrid(null); // Kill AI
				}
				mp[i][j] = null; // Kill ME
			}
		}
	}
	/**
	 * Draws the components of this
	 * 
	 * @param g the Graphics
	 */
	public synchronized void paint(Graphics g)
	{
		this.g = g;
		super.paint(g);
	}
	/**
	 * Overriden to rmove the flickering effect
	 * 
	 * @param g the Graphics
	 */
	public synchronized void update(Graphics g) // Remove Filcker Effect Limits to one the number of enterable MapObjects
	{
		paint(g);
	}
	/**
	 * Repaint the Location specified.
	 * 
	 * @param i the row
	 * @param j the column
	 */
	public synchronized void repaint(int i, int j)
	{
		if(isValidLocation(i, j))
		{
			int w = getSize().width/numj;
			int h = getSize().height/numi;
			int x = j*w/numj;
			int y = i*h/numi;
			g.setClip(x, y, w, h);
			paintAll(g);
		}
	}

	/**
	 * Add a class to be notified of Battles
	 * 
	 * @param the class to notify
	 */
	public void addBattleListener(BattleListener bl)
	{
		if ( battleListeners == null )
			battleListeners = new Vector();
		battleListeners.addElement( bl );
	}
	/**
	 * Removes a class that was being notified.
	 * 
	 * @param the class to stop notifying
	 */
	public void removeBattleListener(BattleListener bl)
	{
			if ( battleListeners != null )
				battleListeners.removeElement( bl );
	}
	/**
	 * Fire a BattleEvent, and inform all classes that are listening
	 * 
	 * @param mp the MapObject of the Players Opponent
	 */
	private synchronized void fireEvent(MapObject mp) 
	{
		if ( battleListeners == null )
			return;
		BattleEvent event = new BattleEvent(this, BattleEvent.BATTLE_STARTED, this.mp[pi][pj], mp);
		for (Enumeration e = battleListeners.elements(); e.hasMoreElements(); )
			((BattleListener)e.nextElement()).battleBegin( event );
	}
	/**
	 * Add a class to be notified of Battles
	 * 
	 * @param the class to notify
	 */
	public void addRedirectListener(RedirectListener r)
	{
		if ( redirectListeners == null )
			redirectListeners = new Vector();
		redirectListeners.addElement( r );
	}
	/**
	 * Removes a class that was being notified.
	 * 
	 * @param the class to stop notifying
	 */
	public void removeRedirectListener(RedirectListener r)
	{
		if ( redirectListeners != null )
			redirectListeners.removeElement( r );
	}
	/**
	 * Fire a RedirectEvent, and inform all classes that are listening
	 * 
	 * @param mp the Redirect
	 * @see Redirect
	 */
	private synchronized void fireRedirectEvent(Redirect mp) 
	{
		synchronized(this.mp[pi][pj])
		{
			battle = true;
			if ( redirectListeners == null )
				return;
			RedirectEvent event = new RedirectEvent(this, RedirectEvent.REDIRECT_FIRED, mp.getMapTo(), (Player)this.mp[pi][pj]);
			for (Enumeration e = redirectListeners.elements(); e.hasMoreElements(); )
				((RedirectListener)e.nextElement()).redirectFired( event );
		}
	}
	/**
	 * Call to begin a battle with th Player, fires the events.
	 * 
	 * @param opp the Opponent of the Player.
	 */
	public synchronized void beginBattleWithPlayer(MapObject opp)
	{
		synchronized(mp[pi][pj]) // Don't let others look at me!
		{							// They're all gonna laugh at me!!
			synchronized(opp)
			{
				if(((Player)mp[pi][pj]).isBattling())
					return;
				battle = true;
				((Player)mp[pi][pj]).battle(); // I'll teach them a lesson!
				fireEvent(opp); // Let's fight, we'll see whose laughing!
			}
		}
	}
}