import java.awt.*;
import java.applet.*;
import javax.swing.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.net.*;
import java.util.*;
/**
 * Class to define People, the superclass of Person, Player and Trainer.  Carry Items
 * Monsters and Monsters at Home
 * 
 * @see Monster
 * @see Item
 * @see Person
 * @see Player
 * @see Trainer
 */
public abstract class People extends MapObject
{
	/**
	 * The Maximum number of Monsters to cary
	 */
	protected final int MAXMON = 4;

	/**
	 * The Monsters caught, and able to be used, ie in the Pepole's hand
	 */
	protected Vector mon; // For the Monsters
	/**
	 * The Monsters caught, and not able to be used, ie in the Pepole's home collection
	 */
	protected Vector hmon; // Monsters at home
	/**
	 * The Items bought.
	 */
	protected Vector items; // For the items

	/**
	 * The index of the active Monster
	 */
	protected int activeMon; // The index of the active Monster
	/**
	 * True if has beed beaten by the Player
	 */
	protected boolean beaten; // Have I been Beaten by player?
	/**
	 * True if in a battle, with Player or Person
	 */
	protected boolean battling; // Am I in a battle?
	/**
	 * The amount of money carrying
	 */
	protected int money; // Amount of cash
	/**
	 * The MapObject to paint this on top of
	 */
	protected static MapObject behind; // Object to Draw Behind
	/**
	 * The direction facing as specified in Grid
	 * 
	 * @see Grid
	 */
	protected int dir; // What way facing
	/**
	 * The name of this
	 */
	protected String name; // The name of the People
	/**
	 * The dialogue to present when in Battle
	 * 
	 * @see Battle
	 */
	protected String dlg;

	/**
	 * Gets the dialogue associated with this
	 * 
	 * @return the dialogue associated with this
	 */
	public String getDialogue() { return dlg; }
	/**
	 * Sets the dialogue associated with this
	 * 
	 * @param slf the dialogue associated with this
	 */
	public void setDialogue(String slf)
	{
		if(slf != null)
			this.dlg = slf;
	}
	/**
	 * Returns false, as this cannot be entered
	 */
	public boolean isEnterable() { return false; }
	/**
	 * Sets the directio of this as specified in Grid
	 * 
	 * @param direct the direction to face
	 * @return true iff direct is a valid direction
	 * @see Grid
	 */
	public boolean setDirection(int direct)
	{
		if(direct == Grid.LEFT || direct == Grid.RIGHT || direct == Grid.UP || direct == Grid.DOWN)
		{
			dir = direct;
			return true;
		}
		else
			return false;
	}
	/**
	 * Gets the direction facing
	 * 
	 * @return the direction facing, as specified in Grid
	 * @see Grid
	 */
	public int getDirection() { return dir; }
	/**
	 * Sets the MapObject to paint behind
	 * 
	 * @param beh the MapObject to paint this over
	 */
	public void setBehind(MapObject beh)
	{
		if(beh.canBeEntered())
		{
			behind = beh;
		}
	}
	/**
	 * Gets the MapObject standing on
	 * 
	 * @return the MapObject standing on
	 */
	public MapObject getBehind() { return behind; }
	/**
	 * Defeat this, ie sets beaten to true.
	 */
	public void beat() { beaten = true; }
	/**
	 * Gets if this is in a Battle
	 * 
	 * @return true if is in a Battle
	 * @see Battle
	 * @see PersonAI
	 */
	public boolean isBattling() { return battling; }
	/**
	 * Begins a battle with this, sets battle to true.
	 */
	public void battle() { battling = true; }
	/**
	 * Ends a battle with this, sets battle to false.
	 */
	public void stopBattle() { battling = false; }
	/**
	 * Sets the amount of Money carried.
	 * 
	 * @param mon the amount of money carried
	 */
	public void setMoney(int mon)
	{
		if(mon > 0) money = mon; 
		else money = 0;
	}
	/**
	 * Gets the amount of Money carried.
	 * 
	 * @return mon the amount of money carried
	 */
	public int getMoney() { return money; }
	/**
	 * Gets the name of this.
	 * 
	 * @return the name of this
	 */
	public String getName() { return name; }
	/**
	 * Sets the name of this.
	 * 
	 * @param name the name of this
	 */
	public void setName(String name)
	{
		if(name != null)
			this.name = name;
	}
	/**
	 * Determine if this has been beaten.
	 * 
	 * @return true iff this has been beaten.
	 */
	public boolean isBeaten() { return beaten; }
	/**
	 * Adds a Monster to this, either to the hand or to 
	 * home if the hand has more than this.MAXMON
	 * 
	 * @param p the Monster to add
	 * @return true if added to hand, false otherwise
	 * @see Monster
	 */
	public boolean addMonster(Monster p)
	{
		if(p == null)
			return false;
		if(mon.size() < MAXMON)
		{
			mon.add(p);
			return true;
		}
		else
		{
			hmon.add(p);
			return false;
		}
	}
	/**
	 * Adds a Monster to the Home collection
	 * 
	 * @param the Monster to add
	 * @see Monster
	 */
	public void addMonsterHome(Monster p)
	{
		if(p != null)
		{
			hmon.add(p);
		}
	}
	/**
	 * Switches a Monster from Home collection to the 
	 * Hand collection.
	 * 
	 * @param hand the Monster in the hand to send home
	 * @param home the Monster at home to send to the hand
	 * @see Monster
	 */
	public boolean switchMonster(Monster hand, Monster home)
	{
		int i, j;
		i = mon.indexOf(hand);
		j = hmon.indexOf(home);
		if(i >= 0 && j >= 0)
		{
			hmon.remove(home);
			mon.remove(hand);
			hmon.add(hand);
			mon.add(home);
			return true;
		}
		return false;
	}
	/**
	 * Sets the ActiveMonster for this.  Does not count Monsters that
	 * have no HP.
	 * 
	 * @param i the new index of the ActiveMonster for this.
	 */
	public void setActiveMonster(int i)
	{
		if(i >= 0 && i < mon.size())
		{
			activeMon = i;
		}
	}
	/**
	 * Gets the index of the Active Monster.  Does not count Monsters that
	 * have no HP.
	 * 
	 * @return the index of the Active Monster
	 */
	public int getActiveMonster() { return activeMon; }
	/**
	 * Gets the Monster from the hand at the specified index, ignoring Monsters that have no HP.
	 * 
	 * @param i the index to retrieve the Monster from
	 * @return the Monster at the specified index, not counting Monsters that have no HP
	 */
	public Monster getMonster(int i)
	{
		int j = 0;
		if(i >= 0 && i < mon.size())
		{
			Monster m;
			for(int q = 0; q < mon.size(); q++)
			{
				m = (Monster)mon.get(q);
				if(m.getHP() > 0)
				{
					if(j == i)
						return m;
					j++;
				}
			}
		}
		return null;
	}
	/**
	 * Gets the Monster from home at the specified index, ignoring Monsters that have no HP.
	 * 
	 * @param i the index to retrieve the Monster from
	 * @return the Monster at the specified index, not counting Monsters that have no HP
	 */
	public Monster getMonsterAtHome(int i)
	{
		int j = 0;
		if(i >= 0 && i < hmon.size())
		{
			Monster m;
			for(int q = 0; q < hmon.size(); q++)
			{
				m = (Monster)hmon.get(q);
				if(m.getHP() > 0)
				{
					if(j == i)
						return m;
					j++;
				}
			}
		}
		return null;
	}
	/**
	 * Gets the number of Monsters in the hand that do not 
	 * have 0 hp.
	 * 
	 * @return the number of Monsters in the hand that do not have 0 hp.
	 */
	public int getNumberOfMonsters()
	{
		int num = 0;
		for(int i = 0; i < mon.size(); i++)
		{
			if(((Monster)mon.get(i)).getHP() > 0)
				num++;
		}
		return num;
	}
	/**
	 * Gets the number of Monsters at home that do not 
	 * have 0 hp.
	 * 
	 * @return the number of Monsters at home that do not have 0 hp.
	 */
	public int getNumberOfMonstersAtHome()
	{
		int num = 0;
		for(int i = 0; i < hmon.size(); i++)
		{
			if(((Monster)hmon.get(i)).getHP() > 0)
				num++;
		}
		return num;	}
	/**
	 * Adds the Item specified to this.
	 * 
	 * @param i the Item to add
	 * @see Item
	 */
	public void addItem(Item i)
	{
		if(i == null)
			return;
		Item it;
		for(int j = 0; j < items.size(); j++)
		{
			it = (Item)items.get(j);
			if(i.equals(it))
			{
				it.addItem();
				return;
			}
		}
		items.add(i);
	}

	/**
	 * Gets the Item corresponding to the name specified.
	 * 
	 * @param the name to search Item Collection for
	 * @return the Item corresponding to the name specified.
	 * @see Item
	 */
	public Item getItem(String name)
	{
		Item it;
		for(int i = 0; i < items.size(); i++)
		{
			it = (Item)items.get(i);
			if(it.getName().equalsIgnoreCase(name))
			{
				return it;
			}
		}
		return null;
	}
	/**
	 * Uses the Item specified on the Monster specified
	 * 
	 * @param it the Item to use on the Monster
	 * @param m the Monster to use the Item on
	 * @see Monster
	 * @see Item
	 */
	public void useItem(Item it, Monster m)
	{
		it.useItem(m);
		if(it.getItemCount() == 0)
			items.remove(it);
	}
	/**
	 * Gets the number of Item this has.
	 * 
	 * @param name the name of the Item to get the count of.
	 * @see Item
	 */
	public int getItemCount(String name)
	{
		Item it;
		for(int i = 0; i < items.size(); i++)
		{
			it = (Item)items.get(i);
			if(it.getName().equalsIgnoreCase(name))
			{
				return it.getItemCount();
			}
		}
		return 0;
	}
	/**
	 * Heals all Monster to full HP, Attack Points, and Defense Points
	 * 
	 * @see Monster
	 * @see Shop
	 */
	public void healAllMonsters()
	{
		Monster m;
		for(int i = 0; i < mon.size(); i++)
		{
			m = (Monster)mon.get(i);
			m.setHP(m.getMaxHP());
			m.setAttack(m.getMaxAttack());
			m.setDefense(m.getMaxDefense());
		}
	}
	/**
	 * Gets an array of String for the Attacks available for the
	 * Monster at index p, ignoring the Monsters with no HP.
	 * 
	 * @param the index of the Monster to get the Attacks for, ignoring those with no HP
	 * @return an array of String for the Attacks available for the Monster at index p, ignoring the Monsters with no HP.
	 * @see Monster
	 * @see Battle
	 */
	public String[] getMonsterAttacks(int p)
	{
		if(p < mon.size() && p >= 0)
		{
			int j = 0;
			Monster m;
			for(int q = 0; q < mon.size(); q++)
			{
				m = (Monster)mon.get(q);
				if(m.getHP() > 0)
				{
					if(j == p)
						return m.getAttacks();
					j++;
				}
			}
		}
		return new String[Monster.MAXATTACKS];
	}
	/**
	 * Gets an array of Strings of the Monsters that the Player is currently carrying, ignoring
	 * those with no HP.
	 * 
	 * @return an array of Strings of the Monsters that the Player is currently carrying, ignoring those with no HP.
	 * @see Monster
	 * @see Battle
	 */
	public String[] getMonsters()
	{
		String[] s = new String[MAXMON];
		Monster m;
		for(int i = 0, j = 0; i < mon.size(); i++)
		{
			m = (Monster)mon.get(i);
			if(m.getHP() > 0)
			{
				s[j++] = m.getName();
			}
		}
		return s;
	}
	/**
	 * Gets an array of Strings of the Item that the Player is currently carrying
	 * 
	 * @return an array of Strings of the Item that the Player is currently carrying.
	 * @see Item
	 * @see Battle
	 * @see Shop
	 */
	public String[] getItems()
	{
		if(items.size() == 0)
			return new String[4];
		String[] s;
		if(items.size() >= 4)
			s = new String[items.size()];
		else
			s = new String[4];
		for(int i = 0; i < items.size(); i++)
		{
			s[i] = ((Item)items.get(i)).getName();
		}
		return s;
	}
}